"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const googleapis_1 = require("googleapis");
const dotenv_1 = require("dotenv");
const async_parallel_1 = require("async-parallel");
const node_ical_1 = __importDefault(require("node-ical"));
const moment_timezone_1 = __importDefault(require("moment-timezone"));
dotenv_1.config();
const oauth2Client = new googleapis_1.google.auth.OAuth2(process.env.GOOGLE_CALENDAR_CLIENT_ID, process.env.GOOGLE_CALENDAR_CLIENT_SECRET, process.env.GOOGLE_CALENDAR_REDIRECT_URL ??
    "https://developers.google.com/oauthplayground");
const globalCalendar = googleapis_1.google.calendar("v3");
/**
 * Split an array in equal chunks
 * @source https://stackoverflow.com/a/51514813/1656944
 */
const chunkArrayInGroups = (arr, parts) => {
    const arrayParts = [];
    for (let i = parts; i > 0; i--)
        arrayParts.push(arr.splice(0, Math.ceil(arr.length / i)));
    return arrayParts;
};
/**
 * Get a random element from this array, strategy-weighted
 */
exports.weightedItemFromArray = (params, arr, strategies) => {
    const originalArray = [...arr];
    const addExtraWeight = (weighter) => {
        originalArray.forEach((item, index) => {
            for (let i = 0; i <
                Math.round(weighter(originalArray, index) * ((params.weight ?? 2) - 1)); i++)
                arr.push(item);
        });
    };
    if (strategies.includes("heavy-firsts"))
        addExtraWeight((array, index) => (array.length - index) / array.length);
    if (strategies.includes("heavy-lasts"))
        addExtraWeight((array, index) => index / array.length);
    if (strategies.includes("heavy-centers"))
        addExtraWeight((array, index) => 1 - Math.abs(array.length / 2 - index) / (array.length / 2));
    [
        "sunday",
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
    ].forEach((day, dayIndex) => {
        if (strategies.includes(`heavy-${day}s`) &&
            params.daily?.timezone)
            addExtraWeight((array, index) => moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").day() ===
                dayIndex
                ? 1
                : 0);
    });
    if (strategies.includes("heavy-mornings") && params.daily?.timezone)
        addExtraWeight((array, index) => moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").hours() > 5 &&
            moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").hours() < 12
            ? 1
            : 0);
    if (strategies.includes("heavy-afternoons") && params.daily?.timezone)
        addExtraWeight((array, index) => moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").hours() >
            11 &&
            moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").hours() < 4
            ? 1
            : 0);
    if (strategies.includes("heavy-evenings") && params.daily?.timezone)
        addExtraWeight((array, index) => moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").hours() > 3 &&
            moment_timezone_1.default.tz(array[index].start, params.daily?.timezone ?? "").hours() < 6
            ? 1
            : 0);
    return arr[Math.floor(Math.random() * arr.length)];
};
/**
 * Live logging for debugging
 */
const log = (params, ...args) => {
    if (params.log) {
        if (typeof params.logger === "function")
            params.logger(...args);
        else
            console.log(args);
    }
};
/**
 * Find a user's events from a single calendar
 * Defaults to their primary calendar
 */
exports.getEventsFromSingleCalendar = async ({ user, from, to, calendar, calendarId, auth, }) => {
    oauth2Client.setCredentials({
        access_token: user?.accessToken ?? process.env.GOOGLE_CALENDAR_ACCESS,
        refresh_token: user?.refreshToken ?? process.env.GOOGLE_CALENDAR_REFRESH,
    });
    return ((await (calendar ?? globalCalendar).events.list({
        timeMin: from ? moment_timezone_1.default(from).toISOString() : undefined,
        timeMax: to ? moment_timezone_1.default(to).toISOString() : undefined,
        auth: auth ?? oauth2Client,
        calendarId: calendarId ?? "primary",
        maxResults: 3,
    })).data.items || []);
};
/**
 * Get a list of user's events from all calendars
 */
exports.getEventsFromAllCalendars = async (params) => {
    oauth2Client.setCredentials({
        access_token: params.user?.accessToken ?? process.env.GOOGLE_CALENDAR_ACCESS,
        refresh_token: params.user?.refreshToken ?? process.env.GOOGLE_CALENDAR_REFRESH,
    });
    const auth = params.auth ?? oauth2Client;
    const calendars = await (params.calendar ?? globalCalendar).calendarList.list({ auth });
    const allEvents = [];
    await async_parallel_1.each(calendars.data.items ?? [], async (cal) => {
        if (cal.id) {
            const events = await exports.getEventsFromSingleCalendar({
                ...params,
                calendarId: cal.id,
            });
            allEvents.push(...events);
        }
    }, { concurrency: 5 });
    return allEvents;
};
/**
 * Get calendar events from a webcal ICS URL
 */
exports.getEventsFromWebcal = async (url) => {
    const directEvents = await node_ical_1.default.async.fromURL(url.replace("webcal:", "http:"));
    return Object.values(directEvents)
        .filter((i) => i.type === "VEVENT")
        .map((i) => {
        const event = i;
        return {
            start: {
                dateTime: event.start.toISOString(),
            },
            end: {
                dateTime: event.end.toISOString(),
            },
        };
    });
};
/**
 * List all free slots for a user
 */
exports.getSlots = async (params) => {
    // Default slot duration is 30 minutes
    const slotDuration = params.slotDuration ?? 30;
    const daysAllowed = params.days ?? [0, 1, 2, 3, 4, 5, 6];
    const strategies = params.strategies ?? ["linear"];
    delete params.slotDuration;
    // Find all slots
    const allPotentialSlots = [];
    const differenceInMinutes = moment_timezone_1.default(params.to).diff(params.from, "minutes");
    let endDate = moment_timezone_1.default(params.from).hours(0).minutes(0).seconds(0);
    while (moment_timezone_1.default(endDate).isBefore(params.to)) {
        const start = endDate;
        const end = moment_timezone_1.default(start).add(slotDuration, "minutes");
        const now = moment_timezone_1.default();
        if (daysAllowed.includes(moment_timezone_1.default(start).day()) &&
            daysAllowed.includes(end.day()) &&
            start.isAfter(moment_timezone_1.default(params.from)
                .hours((params.daily?.from ?? [])[0] ?? 0)
                .minutes((params.daily?.from ?? [])[1] ?? 0)
                .seconds((params.daily?.from ?? [])[2] ?? 0)) &&
            end.isBefore(moment_timezone_1.default(params.to)
                .hours((params.daily?.to ?? [])[0] ?? 0)
                .minutes((params.daily?.to ?? [])[1] ?? 0)
                .seconds((params.daily?.to ?? [])[2] ?? 0)) &&
            start.isAfter(now)) {
            const startTime = moment_timezone_1.default(start);
            const endTime = moment_timezone_1.default(end);
            let dailyConditionsMet = true;
            if (params.daily)
                if (moment_timezone_1.default.tz(startTime, params.daily.timezone).isBefore(moment_timezone_1.default
                    .tz(startTime, params.daily.timezone)
                    .hours((params.daily.from ?? [])[0] ?? 0)
                    .minutes((params.daily.from ?? [])[1] ?? 0)
                    .seconds((params.daily.from ?? [])[2] ?? 0)) ||
                    moment_timezone_1.default.tz(endTime, params.daily.timezone).isAfter(moment_timezone_1.default
                        .tz(endTime, params.daily.timezone)
                        .hours((params.daily.to ?? [])[0] ?? 0)
                        .minutes((params.daily.to ?? [])[1] ?? 0)
                        .seconds((params.daily.to ?? [])[2] ?? 0)))
                    dailyConditionsMet = false;
            if (dailyConditionsMet)
                allPotentialSlots.push({
                    start: startTime.toDate(),
                    end: endTime.toDate(),
                });
        }
        endDate = end;
    }
    log(params, `Total potential slots are ${allPotentialSlots.length}`);
    if (!allPotentialSlots.length)
        return [];
    let calendarEvents = [];
    let timer = new Date().getTime();
    if (params.user || params.auth || params.calendarId || params.calendar) {
        if (params.calendarId)
            calendarEvents = await exports.getEventsFromSingleCalendar(params);
        else
            calendarEvents = await exports.getEventsFromAllCalendars(params);
        log(params, `Fetched ${calendarEvents.length} events from ${params.calendarId ?? "all calendars"} in ${(new Date().getTime() - timer) / 1000} seconds`);
    }
    else if (params.url) {
        calendarEvents = await exports.getEventsFromWebcal(params.url);
        log(params, `Fetched ${calendarEvents.length} events from Webcal`);
    }
    else {
        log(params, "Skipped fetching calendar events");
    }
    let recommendedSlots = allPotentialSlots.filter((slot) => {
        let conflict = false;
        calendarEvents.forEach((event) => {
            if (event.start?.dateTime &&
                event.end?.dateTime &&
                moment_timezone_1.default(slot.start)
                    .subtract(params.padding ?? 0, "minutes")
                    .isAfter(event.start.dateTime) &&
                moment_timezone_1.default(slot.end)
                    .add(params.padding ?? 0, "minutes")
                    .isBefore(event.end.dateTime))
                conflict = true;
        });
        return !conflict;
    });
    if (typeof params.slotFilter === "function")
        recommendedSlots = recommendedSlots.filter(params.slotFilter);
    if (params.slots) {
        if (recommendedSlots.length <= params.slots)
            return recommendedSlots;
        const parts = chunkArrayInGroups(recommendedSlots, params.slots);
        const randomSlots = [];
        for (let i = 0; i < params.slots; i++) {
            let hasSlot = true;
            let slot = exports.weightedItemFromArray(params, parts[i], strategies);
            while (hasSlot) {
                slot = exports.weightedItemFromArray(params, parts[i], strategies);
                hasSlot =
                    randomSlots.find((item) => item.start.getTime() === slot.start.getTime() &&
                        item.end.getTime() === slot.end.getTime()) !== undefined;
            }
            randomSlots.push(slot);
        }
        log(params, `Recommending ${randomSlots.length} slots`);
        return randomSlots;
    }
    return recommendedSlots;
};
//# sourceMappingURL=index.js.map